# -*- coding: utf-8 -*-

from __future__ import absolute_import

import sys
from warnings import warn


class ThriftParserError(Exception):
    pass


class ThriftLexerError(ThriftParserError):
    pass


class ThriftGrammarError(ThriftParserError):
    pass


if sys.version_info >= (3, 7):
    def __getattr__(name):
        if name == "ThriftGrammerError":
            warn("'ThriftGrammerError' is a typo of 'ThriftGrammarError'", DeprecationWarning)
            return ThriftGrammarError

        raise AttributeError("module %r has no attribute %r" % (__name__, name))
else:
    ThriftGrammerError = ThriftGrammarError



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import sys

from .parser import load_module


# TODO: The load process does not compatible with Python standard, e.g., if the
# specified thrift file does not exists, it raises FileNotFoundError, and skiped
# the other meta finders in the sys.meta_path.
if sys.version_info >= (3, 4):
    import importlib.abc
    import importlib.util

    class ThriftImporter(importlib.abc.MetaPathFinder):
        def __init__(self, extension="_thrift"):
            self.extension = extension

        def find_spec(self, fullname, path, target=None):
            if not fullname.endswith(self.extension):
                return None
            return importlib.util.spec_from_loader(fullname,
                                                   ThriftLoader(fullname))


    class ThriftLoader(importlib.abc.Loader):
        def __init__(self, fullname):
            self.fullname = fullname

        def create_module(self, spec):
            return load_module(self.fullname)

        def exec_module(self, module):
            pass
else:
    class ThriftImporter(object):
        def __init__(self, extension="_thrift"):
            self.extension = extension

        def find_module(self, fullname, path=None):
            if fullname.endswith(self.extension):
                return self

        def load_module(self, fullname):
            return load_module(fullname)


_imp = ThriftImporter()


def install_import_hook():
    global _imp
    sys.meta_path[:] = [x for x in sys.meta_path if _imp is not x] + [_imp]


def remove_import_hook():
    global _imp
    sys.meta_path[:] = [x for x in sys.meta_path if _imp is not x]



# -*- coding: utf-8 -*-

"""
Transport for json protocol that apache thrift files will understand
unfortunately, thriftpy2's TJSONProtocol is not compatible with apache's
"""

from __future__ import absolute_import
import json
import base64
import sys

import six

from thriftpy2.protocol import TProtocolBase
from thriftpy2.thrift import TType


CTYPES = {
    TType.BOOL: 'tf',
    TType.BYTE: 'i8',
    TType.I16: 'i16',
    TType.I32: 'i32',
    TType.I64: 'i64',
    TType.DOUBLE: 'dbl',
    TType.STRING: 'str',
    TType.BINARY: 'str',  # apache sends binary data as base64 encoded
    TType.STRUCT: 'rec',
    TType.LIST: 'lst',
    TType.SET: 'set',
    TType.MAP: 'map',
}

JTYPES = {v: k for k, v in CTYPES.items()}

VERSION = 1


def flatten(suitable_for_isinstance):
    """
    isinstance() can accept a bunch of really annoying different types:
        * a single type
        * a tuple of types
        * an arbitrary nested tree of tuples
    Return a flattened tuple of the given argument.
    """

    types = list()

    if not isinstance(suitable_for_isinstance, tuple):
        suitable_for_isinstance = (suitable_for_isinstance,)
    for thing in suitable_for_isinstance:
        if isinstance(thing, tuple):
            types.extend(flatten(thing))
        else:
            types.append(thing)
    return tuple(types)


def _ensure_b64_encode(val):
    """
    Ensure that the variable is something that we can encode with b64encode
    python3 needs bytes, python2 needs string
    """
    if sys.version_info[0] > 2 and isinstance(val, str):
        return val.encode()
    return val


class TApacheJSONProtocolFactory(object):
    @staticmethod
    def get_protocol(trans):
        return TApacheJSONProtocol(trans)


class TApacheJSONProtocol(TProtocolBase):
    """
    Protocol that implements the Apache JSON Protocol
    """

    def __init__(self, trans):
        TProtocolBase.__init__(self, trans)
        self._req = None

    def _load_data(self):
        data = b""
        l_braces = 0
        in_string = False
        while True:
            # read(sz) will wait until it has read exactly sz bytes,
            # so we must read until we get a balanced json list in absence of knowing
            # how long the json string will be
            if hasattr(self.trans, 'getvalue'):
                try:
                    data = self.trans.getvalue()
                    break
                except Exception:
                    pass
            new_data = self.trans.read(1)
            data += new_data
            if new_data == b'"' and not data.endswith(b'\\"'):
                in_string = not in_string
            if not in_string:
                if new_data == b"[":
                    l_braces += 1
                elif new_data == b"]":
                    l_braces -= 1
            if l_braces == 0:
                break
        if data:
            self._req = json.loads(data.decode('utf8'))
        else:
            self._req = None

    def read_message_begin(self):
        if not self._req:
            self._load_data()
        return self._req[1:4]

    def read_message_end(self):
        pass

    def skip(self, ttype):
        pass

    def write_message_end(self):
        pass

    def write_message_begin(self, name, ttype, seqid):
        self.api = name
        self.ttype = ttype
        self.seqid = seqid

    def write_struct(self, obj):
        """
        Write json to self.trans following apache style jsonification of `obj`

        :param obj: A thriftpy2 object
        :return:
        """
        doc = [VERSION, self.api, self.ttype, self.seqid, self._thrift_to_dict(obj)]
        json_str = json.dumps(doc, separators=(',', ':'))
        self.trans.write(json_str.encode("utf8"))

    def _thrift_to_dict(self, thrift_obj, item_type=None):
        """
        Convert a thriftpy2 into an apache conformant dict, eg:

        >>> {0: {'rec': {1: {'str': "304"}, 14: {'rec': {1: {'lst': ["rec", 0]}}}}}}

        >>> {"0":{"rec":{"1":{"str":"284"},"14":{"rec":{"1":{"lst":
        >>>  ["rec",2,{"1":{"i32":12345.0},"2":{"i32":2.0},"3":{"str":"Testing notifications"},"4":{"tf":1}},
              {"1":{"i32":567809.0},"2":{"i32":2.0},"3":{"str":"Other test"},"4":{"tf":0}}]}}}}}}

        :param thrift_obj: the thing we want to make into a dict
        :param item_type: the type of the item we are to convert
        :return:
        """
        if not hasattr(thrift_obj, 'thrift_spec'):
            # use item_type to render it
            if item_type is not None:
                if isinstance(item_type, tuple) and len(item_type) > 1:
                    to_type = item_type[1]
                    flat_key_val = [TType.STRUCT if hasattr(t, 'thrift_spec') else t for t in flatten(to_type)]
                    if flat_key_val[0] == TType.LIST or isinstance(thrift_obj, list):
                        return [CTYPES[flat_key_val[1]], len(thrift_obj)] + [self._thrift_to_dict(v, to_type[1]) for v
                                                                             in thrift_obj]
                    elif flat_key_val[0] == TType.MAP or isinstance(thrift_obj, dict):
                        if to_type[0] == TType.MAP:
                            key_type = flat_key_val[1]
                            val_type = flat_key_val[2]
                        else:
                            key_type = flat_key_val[0]
                            val_type = flat_key_val[1]
                        return [CTYPES[key_type], CTYPES[val_type], len(thrift_obj), {
                            self._thrift_to_dict(k, key_type):
                                self._thrift_to_dict(v, to_type[1]) for k, v in thrift_obj.items()
                        }]
                    if (to_type == TType.BINARY or item_type[-1] == TType.BINARY) and TType.BINARY != TType.STRING:
                        return base64.b64encode(_ensure_b64_encode(thrift_obj)).decode('ascii')
            if isinstance(thrift_obj, bool):
                return int(thrift_obj)
            if (
                item_type == TType.BINARY
                or (isinstance(item_type, tuple) and item_type[0] == TType.BINARY)
            ) and TType.BINARY != TType.STRING:
                return base64.b64encode(_ensure_b64_encode(thrift_obj)).decode("ascii")
            return thrift_obj
        result = {}
        for field_idx, thrift_spec in thrift_obj.thrift_spec.items():
            ttype, field_name, spec = thrift_spec[:3]
            if isinstance(spec, int):
                spec = (spec,)
            val = getattr(thrift_obj, field_name)
            if val is not None:
                if ttype == TType.STRUCT:
                    result[field_idx] = {
                        CTYPES[ttype]: self._thrift_to_dict(val)
                    }
                elif ttype in [TType.LIST, TType.SET]:
                    # format is [list_item_type, length, items]
                    result[field_idx] = {
                        CTYPES[ttype]: [CTYPES[spec[0]], len(val)] + [self._thrift_to_dict(v, spec) for v in val]
                    }
                elif ttype == TType.MAP:
                    key_type = CTYPES[spec[0]]
                    val_type = CTYPES[spec[1][0] if isinstance(spec[1], tuple) else spec[1]]
                    # format is [key_type, value_type, length, dict]
                    result[field_idx] = {
                        CTYPES[ttype]: [key_type, val_type, len(val),
                                        {self._thrift_to_dict(k, spec[0]):
                                         self._thrift_to_dict(v, spec) for k, v in val.items()}]
                    }
                elif ttype == TType.BINARY and TType.BINARY != TType.STRING:
                    result[field_idx] = {
                        CTYPES[ttype]: base64.b64encode(_ensure_b64_encode(val)).decode('ascii')
                    }
                elif ttype == TType.BOOL:
                    result[field_idx] = {
                        CTYPES[ttype]: int(val)
                    }
                else:
                    result[field_idx] = {
                        CTYPES[ttype]: val
                    }
        return result

    def _dict_to_thrift(self, data, base_type):
        """
        Convert an apache thrift dict (where key is the type, value is the data)

        :param data: the dict data
        :param base_type: the type we are going to convert data to
        :return:
        """
        # if the result is a python type, return it:
        if isinstance(data, (str, int, float, bool, six.string_types, six.binary_type)) or data is None:
            if base_type in (TType.I08, TType.I16, TType.I32, TType.I64):
                return int(data)
            if base_type == TType.BINARY and TType.BINARY != TType.STRING:
                return base64.b64decode(data)
            if base_type == TType.BOOL:
                return {
                    'true': True,
                    'false': False,
                    '1': True,
                    '0': False
                }[data.lower()]
            if isinstance(data, bool):
                return int(data)
            return data

        if isinstance(base_type, tuple):
            container_type = base_type[0]
            item_type = base_type[1]
            if container_type == TType.STRUCT:
                return self._dict_to_thrift(data, item_type)
            elif container_type in (TType.LIST, TType.SET):
                return [self._dict_to_thrift(v, item_type) for v in data[2:]]
            elif container_type == TType.MAP:
                return {
                    self._dict_to_thrift(k, item_type[0]):
                        self._dict_to_thrift(v, item_type[1]) for k, v in data[3].items()
                }
        result = {}
        base_spec = base_type.thrift_spec
        for field_idx, val in data.items():
            thrift_spec = base_spec[int(field_idx)]
            # spec has field type, field name, (sub spec), False
            field_name = thrift_spec[1]
            for ftype, value in val.items():
                ttype = JTYPES[ftype]
                if thrift_spec[0] == TType.BINARY and TType.BINARY != TType.STRING:
                    bin_data = val.get('str', '')
                    m = len(bin_data) % 4
                    if m != 0:
                        bin_data += '=' * (4-m)
                    result[field_name] = base64.b64decode(bin_data)

                elif ttype == TType.STRUCT:
                    result[field_name] = self._dict_to_thrift(value, thrift_spec[2])
                elif ttype in (TType.LIST, TType.SET):
                    result[field_name] = [self._dict_to_thrift(v, thrift_spec[2]) for v in value[2:]]
                elif ttype == TType.MAP:
                    key_spec = thrift_spec[2][0]
                    val_spec = thrift_spec[2][1]
                    result[field_name] = {
                        self._dict_to_thrift(k, key_spec): self._dict_to_thrift(v, val_spec)
                        for k, v in value[3].items()
                    }
                else:
                    result[field_name] = {
                        'tf': bool,
                        'i8': int,
                        'i16': int,
                        'i32': int,
                        'i64': int,
                        'dbl': float,
                        'str': str,
                    }[ftype](value)
        if hasattr(base_type, '__call__'):
            return base_type(**result)
        else:
            for k, v in result.items():
                setattr(base_type, k, v)
            return base_type

    def read_struct(self, obj):
        """
        Read the next struct into obj, usually the argument from an incoming request
        Only really used to read the arguments off a request into whatever we want
        see thriftpy2.thrift.TProcessor.process_in for how this class will be used

        Will turn the contents of self.req[4] into the args of obj,
        ie. self.req[4]["1"] must be rendered into obj.thrift_spec

        :param obj:
        :return:
        """
        return self._dict_to_thrift(self._req[4], obj)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

__all__ = [
    'TAsyncTransportBase',
    'TAsyncBufferedTransport',
    'TAsyncBufferedTransportFactory',
    'TAsyncFramedTransport',
    'TAsyncFramedTransportFactory',
]

from .base import TAsyncTransportBase
from .buffered import TAsyncBufferedTransport, TAsyncBufferedTransportFactory
from .framed import TAsyncFramedTransport, TAsyncFramedTransportFactory



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from thriftpy2.thrift import TType

from thriftpy2.protocol.exc import TProtocolException
from thriftpy2.protocol.binary import (
    VERSION_MASK,
    VERSION_1,
    TYPE_MASK,
    unpack_i8,
    unpack_i16,
    unpack_i32,
    unpack_i64,
    unpack_double,
    write_message_begin,
    write_val
)

from .base import TAsyncProtocolBase

BIN_TYPES = (TType.STRING, TType.BINARY)


async def read_message_begin(inbuf, strict=True):
    sz = unpack_i32(await inbuf.read(4))
    if sz < 0:
        version = sz & VERSION_MASK
        if version != VERSION_1:
            raise TProtocolException(
                type=TProtocolException.BAD_VERSION,
                message='Bad version in read_message_begin: %d' % (sz))
        name_sz = unpack_i32(await inbuf.read(4))
        name = await inbuf.read(name_sz)
        name = name.decode('utf-8')

        type_ = sz & TYPE_MASK
    else:
        if strict:
            raise TProtocolException(type=TProtocolException.BAD_VERSION,
                                     message='No protocol version header')

        name = await inbuf.read(sz)
        type_ = unpack_i8(await inbuf.read(1))

    seqid = unpack_i32(await inbuf.read(4))

    return name, type_, seqid


async def read_field_begin(inbuf):
    f_type = unpack_i8(await inbuf.read(1))
    if f_type == TType.STOP:
        return f_type, 0

    return f_type, unpack_i16(await inbuf.read(2))


async def read_list_begin(inbuf):
    e_type = unpack_i8(await inbuf.read(1))
    sz = unpack_i32(await inbuf.read(4))
    return e_type, sz


async def read_map_begin(inbuf):
    k_type = unpack_i8(await inbuf.read(1))
    v_type = unpack_i8(await inbuf.read(1))
    sz = unpack_i32(await inbuf.read(4))
    return k_type, v_type, sz


async def read_val(inbuf, ttype, spec=None, decode_response=True,
                   strict_decode=False):
    if ttype == TType.BOOL:
        return bool(unpack_i8(await inbuf.read(1)))

    elif ttype == TType.BYTE:
        return unpack_i8(await inbuf.read(1))

    elif ttype == TType.I16:
        return unpack_i16(await inbuf.read(2))

    elif ttype == TType.I32:
        return unpack_i32(await inbuf.read(4))

    elif ttype == TType.I64:
        return unpack_i64(await inbuf.read(8))

    elif ttype == TType.DOUBLE:
        return unpack_double(await inbuf.read(8))

    elif ttype == TType.BINARY:
        sz = unpack_i32(await inbuf.read(4))
        return await inbuf.read(sz)

    elif ttype == TType.STRING:
        sz = unpack_i32(await inbuf.read(4))
        byte_payload = await inbuf.read(sz)

        # Since we cannot tell if we're getting STRING or BINARY
        # if not asked not to decode, try both
        if decode_response:
            try:
                return byte_payload.decode('utf-8')
            except UnicodeDecodeError:
                if strict_decode:
                    raise
        return byte_payload

    elif ttype == TType.SET or ttype == TType.LIST:
        if isinstance(spec, tuple):
            v_type, v_spec = spec[0], spec[1]
        else:
            v_type, v_spec = spec, None

        result = []
        r_type, sz = await read_list_begin(inbuf)
        # the v_type is useless here since we already get it from spec
        if (r_type != v_type
                and not (r_type in BIN_TYPES and v_type in BIN_TYPES)):
            for _ in range(sz):
                await skip(inbuf, r_type)
            return []

        for i in range(sz):
            result.append(
                await read_val(inbuf, v_type, v_spec, decode_response,
                               strict_decode)
            )
        return result

    elif ttype == TType.MAP:
        if isinstance(spec[0], int):
            k_type = spec[0]
            k_spec = None
        else:
            k_type, k_spec = spec[0]

        if isinstance(spec[1], int):
            v_type = spec[1]
            v_spec = None
        else:
            v_type, v_spec = spec[1]

        result = {}
        sk_type, sv_type, sz = await read_map_begin(inbuf)
        if sk_type in BIN_TYPES:
            sk_type = k_type
        if sv_type in BIN_TYPES:
            sv_type = v_type
        if sk_type != k_type or sv_type != v_type:
            for _ in range(sz):
                await skip(inbuf, sk_type)
                await skip(inbuf, sv_type)
            return {}

        for i in range(sz):
            k_val = await read_val(inbuf, k_type, k_spec, decode_response,
                                   strict_decode)
            v_val = await read_val(inbuf, v_type, v_spec, decode_response,
                                   strict_decode)
            result[k_val] = v_val

        return result

    elif ttype == TType.STRUCT:
        obj = spec()
        await read_struct(inbuf, obj, decode_response, strict_decode)
        return obj


async def read_struct(inbuf, obj, decode_response=True, strict_decode=False):
    while True:
        f_type, fid = await read_field_begin(inbuf)
        if f_type == TType.STOP:
            break

        if fid not in obj.thrift_spec:
            await skip(inbuf, f_type)
            continue

        if len(obj.thrift_spec[fid]) == 3:
            sf_type, f_name, f_req = obj.thrift_spec[fid]
            f_container_spec = None
        else:
            sf_type, f_name, f_container_spec, f_req = obj.thrift_spec[fid]

        # it really should equal here. but since we already wasted
        # space storing the duplicate info, let's check it.
        if f_type != sf_type:
            if f_type in BIN_TYPES:
                f_type = sf_type
            else:
                await skip(inbuf, f_type)
                continue

        _buf = await read_val(
            inbuf, f_type, f_container_spec, decode_response, strict_decode)
        setattr(obj, f_name, _buf)


async def skip(inbuf, ftype):
    if ftype == TType.BOOL or ftype == TType.BYTE:
        await inbuf.read(1)

    elif ftype == TType.I16:
        await inbuf.read(2)

    elif ftype == TType.I32:
        await inbuf.read(4)

    elif ftype == TType.I64:
        await inbuf.read(8)

    elif ftype == TType.DOUBLE:
        await inbuf.read(8)

    elif ftype in BIN_TYPES:
        _size = await inbuf.read(4)
        await inbuf.read(unpack_i32(_size))

    elif ftype == TType.SET or ftype == TType.LIST:
        v_type, sz = await read_list_begin(inbuf)
        for i in range(sz):
            await skip(inbuf, v_type)

    elif ftype == TType.MAP:
        k_type, v_type, sz = await read_map_begin(inbuf)
        for i in range(sz):
            await skip(inbuf, k_type)
            await skip(inbuf, v_type)

    elif ftype == TType.STRUCT:
        while True:
            f_type, fid = await read_field_begin(inbuf)
            if f_type == TType.STOP:
                break
            await skip(inbuf, f_type)


class TAsyncBinaryProtocol(TAsyncProtocolBase):
    """Binary implementation of the Thrift protocol driver."""

    def __init__(self, trans,
                 strict_read=True, strict_write=True,
                 decode_response=True, strict_decode=False):
        TAsyncProtocolBase.__init__(self, trans)
        self.strict_read = strict_read
        self.strict_write = strict_write
        self.decode_response = decode_response
        self.strict_decode = strict_decode

    async def skip(self, ttype):
        await skip(self.trans, ttype)

    async def read_message_begin(self):
        api, ttype, seqid = await read_message_begin(
            self.trans, strict=self.strict_read)
        return api, ttype, seqid

    async def read_message_end(self):
        pass

    def write_message_begin(self, name, ttype, seqid):
        write_message_begin(
            self.trans, name, ttype,
            seqid, strict=self.strict_write
        )

    def write_message_end(self):
        pass

    async def read_struct(self, obj):
        return await read_struct(self.trans, obj, self.decode_response,
                                 self.strict_decode)

    def write_struct(self, obj):
        write_val(self.trans, TType.STRUCT, obj)


class TAsyncBinaryProtocolFactory(object):
    def __init__(self, strict_read=True, strict_write=True,
                 decode_response=True, strict_decode=False):
        self.strict_read = strict_read
        self.strict_write = strict_write
        self.decode_response = decode_response
        self.strict_decode = strict_decode

    def get_protocol(self, trans):
        return TAsyncBinaryProtocol(
            trans,
            self.strict_read,
            self.strict_write,
            self.decode_response,
            self.strict_decode,
        )



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from ..thrift import TException


class TProtocolException(TException):
    """Custom Protocol Exception class"""

    UNKNOWN = 0
    INVALID_DATA = 1
    NEGATIVE_SIZE = 2
    SIZE_LIMIT = 3
    BAD_VERSION = 4

    def __init__(self, type=UNKNOWN, message=None):
        self.type = type
        self.message = message

    def __str__(self):
        if self.message:
            return self.message

        if self.type == self.UNKNOWN:
            return 'Unknown protocol exception'
        elif self.type == self.INVALID_DATA:
            return 'Invalid data'
        elif self.type == self.NEGATIVE_SIZE:
            return 'Negative size'
        elif self.type == self.SIZE_LIMIT:
            return 'Size limit'
        elif self.type == self.BAD_VERSION:
            return 'Bad version'
        else:
            return 'Default (unknown) TProtocolException'



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from struct import unpack

from thriftpy2.protocol.exc import TProtocolException
from thriftpy2.thrift import TException, TType
from thriftpy2.protocol.compact import (
    from_zig_zag,
    CompactType,
    TCompactProtocol,
)

from .base import TAsyncProtocolBase

BIN_TYPES = (TType.STRING, TType.BINARY)


async def read_varint(trans):
    result = 0
    shift = 0

    while True:
        x = await trans.read(1)
        byte = ord(x)
        result |= (byte & 0x7f) << shift
        if byte >> 7 == 0:
            return result
        shift += 7


class TAsyncCompactProtocol(TCompactProtocol,  # Inherit all of the writing
                            TAsyncProtocolBase):
    """Compact implementation of the Thrift protocol driver."""
    PROTOCOL_ID = 0x82
    VERSION = 1
    VERSION_MASK = 0x1f
    TYPE_MASK = 0xe0
    TYPE_BITS = 0x07
    TYPE_SHIFT_AMOUNT = 5

    async def _read_size(self):
        result = await read_varint(self.trans)
        if result < 0:
            raise TException("Length < 0")
        return result

    async def read_message_begin(self):
        proto_id = await self._read_ubyte()
        if proto_id != self.PROTOCOL_ID:
            raise TProtocolException(TProtocolException.BAD_VERSION,
                                     'Bad protocol id in the message: %d'
                                     % proto_id)

        ver_type = await self._read_ubyte()
        type = (ver_type >> self.TYPE_SHIFT_AMOUNT) & self.TYPE_BITS
        version = ver_type & self.VERSION_MASK
        if version != self.VERSION:
            raise TProtocolException(TProtocolException.BAD_VERSION,
                                     'Bad version: %d (expect %d)'
                                     % (version, self.VERSION))
        seqid = await read_varint(self.trans)
        name = await self._read_string()
        return name, type, seqid

    async def read_message_end(self):  # TAsyncClient expects coroutine
        assert len(self._structs) == 0

    async def _read_field_begin(self):
        type = await self._read_ubyte()
        if type & 0x0f == TType.STOP:
            return None, 0, 0

        delta = type >> 4
        if delta == 0:
            fid = from_zig_zag(await read_varint(self.trans))
        else:
            fid = self._last_fid + delta
        self._last_fid = fid

        type = type & 0x0f
        if type == CompactType.TRUE:
            self._bool_value = True
        elif type == CompactType.FALSE:
            self._bool_value = False

        return None, self._get_ttype(type), fid

    def _read_field_end(self):
        pass

    def _read_struct_begin(self):
        self._structs.append(self._last_fid)
        self._last_fid = 0

    def _read_struct_end(self):
        self._last_fid = self._structs.pop()

    async def _read_map_begin(self):
        size = await self._read_size()
        types = 0
        if size > 0:
            types = await self._read_ubyte()
        vtype = self._get_ttype(types)
        ktype = self._get_ttype(types >> 4)
        return ktype, vtype, size

    async def _read_collection_begin(self):
        size_type = await self._read_ubyte()
        size = size_type >> 4
        type = self._get_ttype(size_type)
        if size == 15:
            size = await self._read_size()
        return type, size

    def _read_collection_end(self):
        pass

    async def _read_byte(self):
        result, = unpack('!b', await self.trans.read(1))
        return result

    async def _read_ubyte(self):
        result, = unpack('!B', await self.trans.read(1))
        return result

    async def _read_int(self):
        return from_zig_zag(await read_varint(self.trans))

    async def _read_double(self):
        buff = await self.trans.read(8)
        val, = unpack('<d', buff)
        return val

    async def _read_binary(self):
        length = await self._read_size()
        return await self.trans.read(length)

    async def _read_string(self):
        length = await self._read_size()
        byte_payload = await self.trans.read(length)

        if self.decode_response:
            try:
                byte_payload = byte_payload.decode('utf-8')
            except UnicodeDecodeError:
                if self.strict_decode:
                    raise
        return byte_payload

    async def _read_bool(self):
        if self._bool_value is not None:
            result = self._bool_value
            self._bool_value = None
            return result
        return (await self._read_byte()) == CompactType.TRUE

    async def read_struct(self, obj):
        self._read_struct_begin()
        while True:
            fname, ftype, fid = await self._read_field_begin()
            if ftype == TType.STOP:
                break

            if fid not in obj.thrift_spec:
                await self.skip(ftype)
                continue

            try:
                field = obj.thrift_spec[fid]
            except IndexError:
                await self.skip(ftype)
                raise
            else:
                if field is not None and \
                        (ftype == field[0]
                         or (ftype in BIN_TYPES
                             and field[0] in BIN_TYPES)):
                    fname = field[1]
                    fspec = field[2]
                    val = await self._read_val(field[0], fspec)
                    setattr(obj, fname, val)
                else:
                    await self.skip(ftype)
            self._read_field_end()
        self._read_struct_end()

    async def _read_val(self, ttype, spec=None):
        if ttype == TType.BOOL:
            return await self._read_bool()

        elif ttype == TType.BYTE:
            return await self._read_byte()

        elif ttype in (TType.I16, TType.I32, TType.I64):
            return await self._read_int()

        elif ttype == TType.DOUBLE:
            return await self._read_double()

        elif ttype == TType.BINARY:
            return await self._read_binary()

        elif ttype == TType.STRING:
            return await self._read_string()

        elif ttype in (TType.LIST, TType.SET):
            if isinstance(spec, tuple):
                v_type, v_spec = spec[0], spec[1]
            else:
                v_type, v_spec = spec, None
            result = []
            r_type, sz = await self._read_collection_begin()

            for i in range(sz):
                result.append(await self._read_val(v_type, v_spec))

            self._read_collection_end()
            return result

        elif ttype == TType.MAP:
            if isinstance(spec[0], int):
                k_type = spec[0]
                k_spec = None
            else:
                k_type, k_spec = spec[0]

            if isinstance(spec[1], int):
                v_type = spec[1]
                v_spec = None
            else:
                v_type, v_spec = spec[1]

            result = {}
            sk_type, sv_type, sz = await self._read_map_begin()
            if sk_type != k_type or sv_type != v_type:
                for _ in range(sz):
                    await self.skip(sk_type)
                    await self.skip(sv_type)
                self._read_collection_end()
                return {}

            for i in range(sz):
                k_val = await self._read_val(k_type, k_spec)
                v_val = await self._read_val(v_type, v_spec)
                result[k_val] = v_val
            self._read_collection_end()
            return result

        elif ttype == TType.STRUCT:
            obj = spec()
            await self.read_struct(obj)
            return obj

    async def skip(self, ttype):
        if ttype == TType.STOP:
            return

        elif ttype == TType.BOOL:
            await self._read_bool()

        elif ttype == TType.BYTE:
            await self._read_byte()

        elif ttype in (TType.I16, TType.I32, TType.I64):
            from_zig_zag(await read_varint(self.trans))

        elif ttype == TType.DOUBLE:
            await self._read_double()

        elif ttype == TType.BINARY:
            await self._read_binary()

        elif ttype == TType.STRING:
            await self._read_string()

        elif ttype == TType.STRUCT:
            self._read_struct_begin()
            while True:
                name, ttype, id = await self._read_field_begin()
                if ttype == TType.STOP:
                    break
                await self.skip(ttype)
                self._read_field_end()
            self._read_struct_end()

        elif ttype == TType.MAP:
            ktype, vtype, size = await self._read_map_begin()
            for i in range(size):
                await self.skip(ktype)
                await self.skip(vtype)
            self._read_collection_end()

        elif ttype == TType.SET:
            etype, size = await self._read_collection_begin()
            for i in range(size):
                await self.skip(etype)
            self._read_collection_end()

        elif ttype == TType.LIST:
            etype, size = await self._read_collection_begin()
            for i in range(size):
                await self.skip(etype)
            self._read_collection_end()


class TAsyncCompactProtocolFactory(object):
    def __init__(self, decode_response=True, strict_decode=False):
        self.decode_response = decode_response
        self.strict_decode = strict_decode

    def get_protocol(self, trans):
        return TAsyncCompactProtocol(
            trans,
            decode_response=self.decode_response,
            strict_decode=self.strict_decode,
        )



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from io import BytesIO

from thriftpy2._compat import CYTHON
from ..base import TTransportBase


class TMemoryBuffer(TTransportBase):
    """Wraps a BytesIO object as a TTransport."""

    def __init__(self, value=None):
        """value -- a value as the initial value in the BytesIO object.

        If value is set, the transport can be read first.
        """
        self._buffer = BytesIO(value) if value is not None else BytesIO()
        self._pos = 0

    def is_open(self):
        return not self._buffer.closed

    def open(self):
        pass

    def close(self):
        self._buffer.close()

    def read(self, sz):
        return self._read(sz)

    def _read(self, sz):
        orig_pos = self._buffer.tell()
        self._buffer.seek(self._pos)
        res = self._buffer.read(sz)
        self._buffer.seek(orig_pos)
        self._pos += len(res)
        return res

    def write(self, buf):
        self._buffer.write(buf)

    def flush(self):
        pass

    def getvalue(self):
        return self._buffer.getvalue()

    def setvalue(self, value):
        self._buffer = BytesIO(value)
        self._pos = 0


if CYTHON:
    from .cymemory import TCyMemoryBuffer  # noqa
